﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  [StorableClass]
  public abstract class Symbol : DeepCloneable, IEquatable<Symbol> {
    [Storable]
    private readonly int stringRepresentationHash;

    [Storable]
    public string StringRepresentation { get; private set; }

    protected Symbol(string representation) {
      StringRepresentation = representation;
      stringRepresentationHash = representation.GetHashCode();
    }

    protected Symbol(Symbol original, Cloner cloner) : base(original, cloner) {
      StringRepresentation = original.StringRepresentation;
      stringRepresentationHash = original.stringRepresentationHash;
    }

    [StorableConstructor]
    protected Symbol(bool deserializing) { }

    public override string ToString() {
      return StringRepresentation;
    }

    #region IEquatable
    public static bool operator ==(Symbol s1, Symbol s2) {
      if (ReferenceEquals(s1, s2)) return true;
      if (ReferenceEquals(s1, null) || ReferenceEquals(s2, null)) return false;
      return s1.Equals(s2);
    }

    public static bool operator !=(Symbol s1, Symbol s2) {
      return !(s1 == s2);
    }

    public bool Equals(Symbol other) {
      if (ReferenceEquals(other, null)) return false;
      if (ReferenceEquals(other, this)) return true;
      if (this.GetType() != other.GetType()) return false; // Otherwise, this needs to be reimplemented in derived classes.
      return StringRepresentation == other.StringRepresentation;
    }

    public override bool Equals(object obj) {
      if (ReferenceEquals(obj, null)) return false;
      if (ReferenceEquals(obj, this)) return true;
      if (this.GetType() != obj.GetType()) return false;
      return Equals((Symbol)obj);
    }

    public override int GetHashCode() {
      return stringRepresentationHash;
    }
    #endregion
  }

  [StorableClass]
  public class TerminalSymbol : Symbol {
    public TerminalSymbol(string representation) : base(representation) { }
    public TerminalSymbol(TerminalSymbol original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TerminalSymbol(this, cloner);
    }

    [StorableConstructor]
    protected TerminalSymbol(bool deserializing) : base(deserializing) { }
  }

  [StorableClass]
  public class VariableTerminalSymbol : TerminalSymbol {
    public VariableTerminalSymbol(string representation) : base(representation) { }
    public VariableTerminalSymbol(VariableTerminalSymbol original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VariableTerminalSymbol(this, cloner);
    }

    [StorableConstructor]
    protected VariableTerminalSymbol(bool deserializing) : base(deserializing) { }
  }

  [StorableClass]
  public class NonterminalSymbol : Symbol {
    public NonterminalSymbol(string representation) : base(representation) { }
    public NonterminalSymbol(NonterminalSymbol original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new NonterminalSymbol(this, cloner);
    }

    [StorableConstructor]
    protected NonterminalSymbol(bool deserializing) : base(deserializing) { }
  }


}
